﻿#region File Description
//-----------------------------------------------------------------------------
// GameBoard.cs
// Contains functions for interacting with the game board.  This includes tower and
// enemy placement, tower actions, enemy movement, etc.
//
// Authors: John Karczewski, Kelsey Schlarman, Eric Barr, Kunal Dalvi, Shajee Fakhi, Will Duff
//-----------------------------------------------------------------------------
#endregion


using System.Collections.Generic;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;
using SpriteSheetRuntime;

namespace ZuneDefense
{
    public class GameBoard : DrawableZuneDefenseComponent
    {
        /// <summary>
        /// Number of tiles in the x direction.
        /// </summary>
        public static readonly int WidthInTiles = 15;

        /// <summary>
        /// Number of tiles in the y direction.
        /// </summary>
        public static readonly int HeightInTiles = 42;

        /// <summary>
        /// Width of a tile in pixels.
        /// </summary>
        public static readonly int TileWidthInPixels = 15;

        /// <summary>
        /// Height of a tile in pixels.
        /// </summary>
        public static readonly int TileHeightInPixels = 15;

        /// <summary>
        /// Number of pixels in the x direction.
        /// </summary>
        public static readonly int WidthInPixels = WidthInTiles * TileWidthInPixels;

        /// <summary>
        /// Number of pixels in the y direction.
        /// </summary>
        public static readonly int HeightInPixels = HeightInTiles * TileHeightInPixels;

        /// <summary>
        /// The width of the border along the left edge of the gameboard.
        /// </summary>
        public static readonly int LeftBoundaryInPixels = 8;

        /// <summary>
        /// The width of the border along the right edge of the gameboard.
        /// </summary>
        public static readonly int RightBoundaryInPixels = 7;

        /// <summary>
        /// The height of the spawn area on the top of the gameboard. 
        /// </summary>
        public static readonly int TopBoundaryInTiles = 3;

        /// <summary>
        /// The height of the exit area on the bottom of the gameboard.
        /// </summary>
        public static readonly int BottomBoundaryInTiles = 3;

        /// <summary>
        /// The height of the spawn area on the top of the gameboard. 
        /// </summary>
        public static readonly int TopBoundaryInPixels = TopBoundaryInTiles * TileHeightInPixels;

        /// <summary>
        /// The height of the exit area on the bottom of the gameboard.
        /// </summary>
        public static readonly int BottomBoundaryInPixels = BottomBoundaryInTiles * TileHeightInPixels;

        public static Tile[,] Tiles = new Tile[WidthInTiles, HeightInTiles];

        private List<Tower> towers;

        /// <summary>
        /// Constructs a new GameBoard.
        /// </summary>
        public GameBoard(SpriteSheet spriteSheet, SpriteBatch spriteBatch)
            : base(spriteSheet, spriteBatch, spriteSheet.GetIndex("GameBackground"), Vector2.Zero, new Rectangle(0, 0, WidthInPixels, HeightInPixels))
        {
            for (int x = 0; x < WidthInTiles; x++)
            {
                for (int y = 0; y < HeightInTiles; y++)
                {
                    Tiles[x, y] = new Tile(spriteSheet, spriteBatch, BoardToScreen(x, y));
                }
            }

            towers = new List<Tower>();
            PlotShortestPaths();
        }

        /// <summary>
        /// Converts a tile coordinate into a pixel coordinate.
        /// </summary>
        public static Vector2 BoardToScreen(int x, int y)
        {
            return new Vector2(LeftBoundaryInPixels + (float)x * 15, 
                (float)y * 15);
        }

        /// <summary>
        /// Returns the tile coordinate containing the given pixel coordinate.
        /// </summary>
        public static Vector2 ScreenToBoard(Vector2 pixelPosition)
        {
            return new Vector2((int)((pixelPosition.X - GameBoard.LeftBoundaryInPixels) / 15),
                (int)(pixelPosition.Y  / 15));
        }

        /// <summary>
        /// Updates the towers and makes sure they continue to target new enemies when old targets go
        /// out of range.
        /// </summary>
        public override void Update(GameTime gameTime)
        {
            SwarmManager swarmManager = GameplayScreen.swarmManager;
            foreach (Tower tower in towers)
            {
                if (!tower.hasEnemy())
                {
                    Enemy closestEnemy = FindClosestEnemy(tower, swarmManager.CurrentSwarms);
                      tower.setEnemy(closestEnemy);
                }
                tower.Update(gameTime);
            }
        }

        /// <summary>
        /// Finds the closest enemy to a given tower.
        /// </summary>
        public Enemy FindClosestEnemy(Tower tower, List<Swarm> swarms)
        {
            double closestEnemyDistance = double.PositiveInfinity;
            Enemy closestEnemy = null;

            foreach (Swarm swarm in swarms)
            {
                foreach (Enemy enemy in swarm.Enemies)
                {
                    if ((enemy.IsAir && !tower.ShootsAir) || (!enemy.IsAir && !tower.ShootsGround))
                    {
                        continue;
                    }

                    double distance = Vector2.Distance(enemy.CenterPixelPosition, tower.CenterPixelPosition);
                    if (distance < closestEnemyDistance)
                    {
                        closestEnemyDistance = distance;
                        closestEnemy = enemy;
                    }
                }
            }

            return closestEnemy;
        }

        /// <summary>
        /// Draws the game board and all child objects.
        /// </summary>
        public override void Draw(GameTime gameTime, Camera camera)
        {
            // Draw the game board background.
            spriteBatch.Draw(spriteSheet.Texture, new Vector2(0, camera.Position.Y), spriteSheet.SourceRectangle(textureIndex), Color.White);

            // Draw tiles.
            foreach (Tile tile in Tiles)
            {
                if (tile.IsOccupied)
                {
                    tile.Draw(gameTime, camera);
                }
            }
        }

        /// <summary>
        /// Spawns a tower at the given tile coordinates. Returns true if the tower was
        /// spawned successfully and false otherwise.
        /// </summary>
        public bool SpawnTower(Tower t)
        {
            int x = (int)t.TilePosition.X;
            int y = (int)t.TilePosition.Y;

            // Check that 4 tiles are not occupied.
            if (CanSpawnTower(x, y))
            {
                if (Tiles[x, y].spawnTower(t))
                {
                    Tiles[x + 1, y].IsOccupied = true;
                    Tiles[x, y + 1].IsOccupied = true;
                    Tiles[x + 1, y + 1].IsOccupied = true;
                    towers.Add(t);

                    PlotShortestPaths();

                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Checks that a tower can be spawned at the given tile coordinates.
        /// </summary>
        public bool CanSpawnTower(int x, int y)
        {
            foreach (Swarm swarm in GameplayScreen.swarmManager.CurrentSwarms)
            {
                foreach (Enemy enemy in swarm.Enemies)
                {
                    if (enemy.CenterTilePosition.X == x && enemy.CenterTilePosition.Y == y
                        || enemy.CenterTilePosition.X == x + 1 && enemy.CenterTilePosition.Y == y
                        || enemy.CenterTilePosition.X == x && enemy.CenterTilePosition.Y == y + 1
                        || enemy.CenterTilePosition.X == x + 1 && enemy.CenterTilePosition.Y == y + 1)
                        return false;
                }
            }

            // Do some bounds checking.
            if (TowerBoundCheck(x, y))
                return false;

            // Check that the 4 tiles are not occupied.
            if (CheckTilesAround(x, y))
                return true;

            return false;
        }

        private static bool CheckTilesAround(int x, int y)
        {
            return !Tiles[x, y].IsOccupied && !Tiles[x + 1, y].IsOccupied && !Tiles[x, y + 1].IsOccupied && !Tiles[x + 1, y + 1].IsOccupied;
        }

        private static bool TowerBoundCheck(int x, int y)
        {
            return x < 0 || x + 1 > WidthInTiles || y < 0 || y + 1 > HeightInTiles;
        }

        /// <summary>
        /// Returns true if a tower is located at the given tile coordinates and false otherwise.
        /// </summary>
        public bool TileContainsTower(int x, int y)
        {
            return (Tiles[x, y].tower != null);
        }

        /// <summary>
        /// Returns the tower at the given tile coordinates.
        /// </summary>
        public Tower GetTower(int x, int y)
        {
            return Tiles[x, y].tower;
        }

        /// <summary>
        /// Removes a tower at the given tile coordinates.
        /// </summary>
        public void RemoveTower(int x, int y)
        {
            if (Tiles[x, y].tower != null)
            {
                towers.Remove(Tiles[x, y].tower);
                Tiles[x, y].tower = null;
                Tiles[x, y].IsOccupied = false;
                Tiles[x + 1, y].IsOccupied = false;
                Tiles[x, y + 1].IsOccupied = false;
                Tiles[x + 1, y + 1].IsOccupied = false;
            }

            PlotShortestPaths();
        }

        /// <summary>
        /// Plots the shortest path from all tiles to the bottom of the gameboard. Each tile holds a pointer
        /// to the next tile in the shortest path to the bottom. An enemy simply follows these pointers to 
        /// find its way to the bottom of the map.
        /// </summary>
        private void PlotShortestPaths()
        {
            Queue<Tile> openSet = new Queue<Tile>();
            
            // Reset all the tiles.
            foreach (Tile tile in Tiles)
            {
                tile.Seen = tile.IsOccupied;
            }
            
            // Add all the bottom tiles into the queue.
            for (int x = 0; x < WidthInTiles; x++)
            {
                openSet.Enqueue(Tiles[x, HeightInTiles - 1]);
            }

            // Find the shortest path for all tiles.
            while (openSet.Count > 0)
            {
                Tile currentTile = openSet.Dequeue();

                if (currentTile.TilePosition.X > 0)
                {
                    //West neighbor.
                    AddNeighbor(openSet, currentTile, GameBoard.Tiles[(int)currentTile.TilePosition.X - 1, (int)currentTile.TilePosition.Y]);
                }
                if (currentTile.TilePosition.X < GameBoard.WidthInTiles - 1)
                {
                    //East neighbor.
                    AddNeighbor(openSet, currentTile, GameBoard.Tiles[(int)currentTile.TilePosition.X + 1, (int)currentTile.TilePosition.Y]);
                }

                if (currentTile.TilePosition.Y > 0)
                {
                    //North neighbor.
                    AddNeighbor(openSet, currentTile, GameBoard.Tiles[(int)currentTile.TilePosition.X, (int)currentTile.TilePosition.Y - 1]);
                }
                if (currentTile.TilePosition.Y < GameBoard.HeightInTiles - 1)
                {
                    //South neighbor.
                    AddNeighbor(openSet, currentTile, GameBoard.Tiles[(int)currentTile.TilePosition.X, (int)currentTile.TilePosition.Y + 1]);
                }
            }
        }

        /// <summary>
        /// Our shortest path algorithm adds all neighbor tiles of the current tile to a queue so they can be 
        /// searched through later.
        /// </summary>
        private void AddNeighbor(Queue<Tile> openSet, Tile currentTile, Tile neighborTile)
        {
            if (!neighborTile.Seen && !neighborTile.IsOccupied)
            {
                neighborTile.Seen = true;
                neighborTile.Next = currentTile;
                openSet.Enqueue(neighborTile);
            }
        }

        /// <summary>
        /// When a tower is blocking the path of one or more enemies, an enemy uses this method to
        /// destroy the tower that lies in its path.
        /// </summary>
        public void FindAndDestroyTower(Tile tileToDestroy)
        {
            Vector2 locationToDestroy = tileToDestroy.TilePosition;

            // Towers can be on one of four tiles.
            if (tileToDestroy.tower == null)
            {
                if (Tiles[(int)tileToDestroy.TilePosition.X - 1, (int)tileToDestroy.TilePosition.Y].tower != null)
                {
                    // West neighbor.
                    locationToDestroy = new Vector2((int)tileToDestroy.TilePosition.X - 1, (int)tileToDestroy.TilePosition.Y);
                }
                else if (Tiles[(int)tileToDestroy.TilePosition.X, (int)tileToDestroy.TilePosition.Y - 1].tower != null)
                {
                    // North neighbor.
                    locationToDestroy = new Vector2((int)tileToDestroy.TilePosition.X, (int)tileToDestroy.TilePosition.Y - 1);
                }
                else if (Tiles[(int)tileToDestroy.TilePosition.X - 1, (int)tileToDestroy.TilePosition.Y - 1].tower != null)
                {
                    // North-west neighbor.
                    locationToDestroy = new Vector2((int)tileToDestroy.TilePosition.X - 1, (int)tileToDestroy.TilePosition.Y - 1);
                }
            }
            
            RemoveTower((int)locationToDestroy.X, (int)locationToDestroy.Y);
        }
    }
}
